/* NFIDENT: a program for neuro-fuzzy function approximation

Copyright (C) 1997-2000 

Detlef Nauck (nauck@ieee.org, 
http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
hereby disclaims all copyright interests in the program 
NFIDENT written by Detlef Nauck

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define VERSION "Version 0.91 Beta, July 10, 2000"

#define sqr(x) ((x)*(x))
#define sgn(x) ((x<0)?-1:(x==0)?0:1)

#define TRIANGLE   0
#define TRAPEZOID  1
#define GAUSSIAN   2
#define MAXFS      15
#define MAXIN      500
#define MAXOUT     MAXIN
#define MAXRULES   500
#define SAMPLES    100
#define TMIN       0
#define TPROD      1
#define SMAX       0

#define DCOG       0
#define DMOM       1

#define MAMDANI    0
#define SUGENO     1

#define GF 1.7 /* factor for controlling overlap of gaussian msf */


#define GNU_TRIANGLE "(x) = ((x>=%f)&&(x<=%f)?((x-%f)/(%f-%f)):(x<=%f)&&(x>=%f)?((%f-x)/(%f-%f)):-1)\n"
/*"(x) = ((x>=a)&&(x<=b)?((x-a)/(b-a)):(x<=c)&&(x>=b)?((c-x)/(c-b)):-1)\n"*/

#define GNU_LSTRIANGLE "(x) = ((x<=%f)?1:(x>%f)&&(x<=%f)?(%f-x)/(%f-%f):-1)\n"
/*"(x) = ((x<=b)?1:(x>b)&&(x<=c)?(c-x)/(c-b):-1)\n"*/

#define GNU_RSTRIANGLE "(x) = ((x>=%f)?1:(x<%f)&&(x>=%f)?(x-%f)/(%f-%f):-1)\n"
/*"(x) = ((x>=b)?1:(x<b)&&(x>=a)?(x-a)/(b-a):-1)\n"*/

#define GNU_TRAPEZOID "(x) = ((x>=%f)&&(x<=%f)?(x-%f)/(%f-%f):(x>%f)&&(x<%f)?1:(x>=%f)&&(x<=%f)?(%f-x)/(%f-%f):-1)\n"
/*"(x) = ((x>=a)&&(x<=b)?(x-a)/(b-a):(x>b)&&(x<c)?1:(x>=c)&&(x<=d)?(d-x)/(d-c):-1)\n"*/

#define GNU_LSTRAPEZOID "(x) = ((x<%f)?1:(x>=%f)&&(x<=%f)?(%f-x)/(%f-%f):-1)\n"
/*"(x) = ((x<c)?1:(x>=c)&&(x<=d)?(d-x)/(d-c):-1)\n"*/

#define GNU_RSTRAPEZOID "(x) = ((x>=%f)&&(x<=%f)?(x-%f)/(%f-%f):(x>%f)?1:-1)\n"
/*"(x) = ((x>=a)&&(x<=b)?(x-a)/(b-a):(x>b)?1:-1)\n"*/

#define GNU_EXPMSF "(x) = exp (- ( %f * (%f - x) / %f) ** 2)\n"
/*"(x) = exp (- ( c * (b - x) / a) ** 2)\n"*/


/* type definitions */

typedef double *matrix;

typedef struct PatternSet 
{
	int count;
	int in;
	int out;
	double *pmax;
	double *pmin;
	double *pmean;
	double *pstddev;
	matrix *Pats;
} PSet;


typedef struct LearnParameters
{
	int fstype;
	int imsf[MAXIN];  
	int omsf[MAXOUT];
	int iform[MAXIN];
	int oform[MAXOUT];
	double outstretch;
	double tskinit;
	double sigma_a;
	double sigma_b;
	double sigma_c;
	double sigma_d;
	double sigma_t;
	double momentum;
	int   uselmse;
	int   tnorm;
	int   snorm;
	int   defuzz;
	int   maxrules;
	double percent;
	double minerror;
	int   chgepoch;
	int   minepoch;
	int   maxepoch;
	int   logepoch;
	int   batch;
	int   at05;
	int   nopass;
	int   asym;
	char  netfilename[80];
	char  loadnetfilename[80];
	char  savenetfilename[80];
	char  logfilename[80];
	char  gnufilename[80];
	char  trainingfilename[80];
	char  generalizationfilename[80];
	char  testfilename[80];
	char  trainresultname[80];
	char  genresultname[80];
	char  testresultname[80];
} LParam;

/* Fuzzy set, a-d are parameters, ls and rs are flags for left and right */
/* shouldered, form denotes the type of membership function, hit is the  */
/* current membership degree or degree of fulfillment, d_a - d_d are     */
/* changes for the parameters during learning.                           */
typedef struct FuzzySet  
{
	char   name[20];
	double  a,b,c,d;
	double  d_a, d_b, d_c, d_d;
	double  od_a, od_b, od_c, od_d;
	double  hit;
	int    ls, rs;
	int    form; /* TRIANGLE, TRAPEZOID or GAUSSIAN */
} FSet;

/* linguistic variable, partition stores a fuzzy partition */
typedef struct LinguisticVariable  
{
	char name[20];
	double upper, lower;            /* upper and lower boundary */
	int count;                     /* number of fuzzy sets */
	int form;                      /* form of fuzzy sets */
	FSet fsets[MAXFS];             /* fuzzy partition */
} LingVar;

typedef LingVar *PLingVar;


/* A Mamdani fuzzy rule; antecedents and conclusion are pointers to arrays */
/* of indeces of fuzzy sets; hit is the degree of fulfillment of the rule, */
/* error is the individual rule error. */
typedef struct MamdaniFuzzyRule  
{
	int *antecedent;
	int *conclusion;
	double hit;
	double error;
} MRule;

typedef MRule *PMRule;

/* A TSK fuzzy rule; antecedents is a pointer to an array of indeces       */
/* of fuzzy sets, conclusion is a pointer array. Each pointer references   */
/* an array of doubles. The arrays store coefficients to be multiplied with */
/* the input values. hit is the degree of fulfillment of the rule, normhit */
/* is the  normalized degree of fulfillment, and error is the individual   */
/* rule error. */
typedef struct TSKFuzzyRule  
{
	int *antecedent;
	matrix *conclusion;
	matrix *delta;
	matrix *old_delta;
	double *output;
	double hit;
	double normhit; 
	double error;
} TSKRule;

typedef TSKRule *PTSKRule;

typedef double (*TNORM) (double, double);
typedef double (*SNORM) (double, double);

typedef double (*DEFUZZ) (LingVar *, TNORM, SNORM);

/* Mamdani type fuzzy system */
typedef struct MamdaniFuzzySystem  
{
	PLingVar *invars;            /* pointer to the input variable pointers  */
	PLingVar *outvars;           /* pointer to the output variable pointers */
	PMRule *rulebase;            /* pointer to rulebase of the fuzzy system */
	int in;                      /* number of input variables */
	int out;                     /* number of output variables */
	int rules;                   /* number of rules in rulebase */
	TNORM tnorm;                 /* pointer to a t-norm  */
	SNORM snorm;                 /* pointer to a s-norm */
	DEFUZZ defuzz;               /* pointer to a defuzzification function  */
	double *outputs;             /* pointer to defuzzified output vector */
	double *errors;              /* pointer to array of output errors */
	double Error;                /* sum of squared errors */
	matrix *meanout;             /* matrix of weighted averages for rule targets */
	matrix *variance;            /* matrix of variances for rule targets */
	double *performance;         /* vector of rule performances */
	double *sumhits;             /* sum of degrees of fulfilment for each rule */
	int *pats;                   /* vector that counts how many patterns each rule covers */
} MFuzzySystem;


/* Takagi-Sugeno-Kang type fuzzy system */
typedef struct TakagiSugenoKangFuzzySystem  
{
	PLingVar *invars;            /* pointer to the input variable pointers */
	PTSKRule *rulebase;          /* pointer to rulebase of the fuzzy system */
	int in;                      /* number of input variables */
	int out;                     /* number of output variables */
	int rules;                   /* number of rules in rulebase  */
	TNORM tnorm;                 /* pointer to t-norm */
	double *outputs;              /* pointer to output vector */
	double *errors;               /* pointer to array of output errors */
	double Error;                 /* sum of squared errors */
} TSKFuzzySystem;


/* global functions */

double minimum (double x, double y);
double prod (double x, double y);

double maximum (double x, double y);
void Init_Learn_Parameters (LParam *LP);
void INTHandler();

/* global variables */

PSet *TrainingPattern;
PSet *GeneralizationPattern;
PSet *TestPattern;
LParam LP;
MFuzzySystem *Nefprox;
MFuzzySystem *BestNefprox;
TSKFuzzySystem *Anfis;
TSKFuzzySystem *BestAnfis;
int StopOnSignal;
