/* NFIDENT: a program for neuro-fuzzy function approximation

   Copyright (C) 1997-2000 

   Detlef Nauck (nauck@ieee.org, 
   http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
   hereby disclaims all copyright interests in the program 
   NFIDENT written by Detlef Nauck

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <string.h>
#include "global.h"
#include "fuzzy.h"
#include "mamdani.h"
#include "defuzz.h"
#include "tsk.h"

void Remove_Whitespace (char *cmdline)
{
  char *S;
  int i;

  S = cmdline; i = 0;
  while (((cmdline[i] == ' ') || (cmdline[i] == '\t') || 
         (cmdline[i] == '\n')) && (cmdline[i] != '\0'))
    {S++; i++;}

  strcpy (cmdline, S);
}; /* Remove_Whitespace */



int Process_Cmd (char *cmdline, LParam *LP)
{
  double value;
  int i;
  char S[80];

  sscanf (cmdline, "%s %lf", S, &value);

  if ((strstr (cmdline, "fstype")) != NULL)
    {
      if ((value >= MAMDANI) && (value <= SUGENO))
	LP->fstype = (int) value;
      return 0;
    };
  
  if ((strstr (cmdline,"imsf[")) != NULL)
    {
      i = atoi (&cmdline[5]);
      if ((i >= 0) && (i <= MAXIN) && (value > 1) &&(value <= 15))
	LP->imsf[i] = (int) value;
      return 0;
    };

  if ((strstr (cmdline, "imsf")) != NULL)
    {
      if ((value > 1) && (value <= 15))
	for (i=0; i < MAXIN; i++)
	  LP->imsf[i] = (int) value;
      return 0;
    };
  
  if ((strstr (cmdline,"iform[")) != NULL)
    {
      i = atoi (&cmdline[6]);
      if ((i >= 0) && (i <= MAXIN) && (value >= TRIANGLE) 
	  && (value <= GAUSSIAN))
	LP->iform[i] = (int) value;
      return 0;
    };

  if ((strstr (cmdline, "iform")) != NULL)
    {
      if ((value >= TRIANGLE) && (value <= GAUSSIAN))
	for (i=0; i < MAXIN; i++)
	  LP->iform[i] = (int) value;
      return 0;
    };
  
  if ((strstr (cmdline,"omsf[")) != NULL)
    {
      i = atoi (&cmdline[5]);
      if ((i >= 0) && (i <= MAXIN) && (value > 1) &&(value <= 15))
	LP->omsf[i] = (int) value;
      return 0;
    };

  if ((strstr (cmdline, "omsf")) != NULL)
    {
      if ((value > 1) && (value <= 15))
	for (i=0; i < MAXIN; i++)
	  LP->omsf[i] = (int) value;
      return 0;
    };

  if ((strstr (cmdline,"oform[")) != NULL)
    {
      i = atoi (&cmdline[6]);
      if ((i >= 0) && (i <= MAXIN) && (value >= TRIANGLE) 
	  && (value <= GAUSSIAN))
	LP->oform[i] = (int) value;
      return 0;
    };

  if ((strstr (cmdline, "oform")) != NULL)
    {
      if ((value >= TRIANGLE) && (value <= GAUSSIAN))
	for (i=0; i < MAXIN; i++)
	  LP->oform[i] = (int) value;
      return 0;
    };

  if ((strstr (cmdline, "outstretch")) != NULL)
    {
      LP->outstretch = value;
      return 0;
    };

  if ((strstr (cmdline, "tnorm")) != NULL)
    {
      if ((value >= TMIN) && (value <= TPROD))
	LP->tnorm = (int) value;
      return 0;
    };
  
  if ((strstr (cmdline, "tconorm")) != NULL)
    {
      if ((value >= SMAX) && (value <= SMAX))
	LP->snorm = (int) value;
      return 0;
    };
     
  if ((strstr (cmdline, "defuzz")) != NULL)
    {
      if ((value >= DCOG) && (value <= DMOM))
	LP->defuzz = (int) value;
      return 0;
    };
  
  if ((strstr (cmdline, "tskinit")) != NULL)
    {
      LP->tskinit = value;
      return 0;
    };
  
  
  if ((strstr (cmdline, "sigma_a")) != NULL)
    {
      LP->sigma_a = value;
      return 0;
    };
  
  if ((strstr (cmdline, "sigma_b")) != NULL)
    {
      LP->sigma_b = value;
      return 0;
    };
  
  if ((strstr (cmdline, "sigma_c")) != NULL)
    {
      LP->sigma_c = value;
      return 0;
    };
  
  if ((strstr (cmdline, "sigma_d")) != NULL)
    {
      LP->sigma_d = value;
      return 0;
    };
 
  if ((strstr (cmdline, "sigma_t")) != NULL)
    {
      LP->sigma_t = value;
      return 0;
    };
  
  if ((strstr (cmdline, "sigma")) != NULL)
    {
      LP->sigma_a = value;
      LP->sigma_b = value;
      LP->sigma_c = value;
      LP->sigma_d = value;
      LP->sigma_t = value;
      return 0;
    };

  if ((strstr (cmdline, "momentum")) != NULL)
    {
      LP->momentum = value;
      return 0;
    };
  
  if ((strstr (cmdline, "uselmse")) != NULL)
    {
      if (!(int) value)
	LP->uselmse = 0;
      else
	LP->uselmse = 1;
      return 0;
    };

  if ((strstr (cmdline, "maxrules")) != NULL)
    {
      if (value >= -1)
	{
	  LP->maxrules = (int) value;
	  return 0;
	}
      else 
	return -1;
    };

  if ((strstr (cmdline, "percent")) != NULL)
    {
      if ((value > 0) && (value <= 1))
	{
	  LP->percent = value;
	  return 0;
	}
      else 
	return -1;
    };
  
  if ((strstr (cmdline, "minerror")) != NULL)
    {
       LP->minerror = value;
      return 0;
    };
  
  if ((strstr (cmdline, "maxepoch")) != NULL)
    {
      LP->maxepoch = (int) value;
      return 0;
    };
  
  if ((strstr (cmdline, "minepoch")) != NULL)
    {
      LP->minepoch = (int) value;
      return 0;
    };
  
  if ((strstr (cmdline, "chgepoch")) != NULL)
    {
      LP->chgepoch = (int) value;
      return 0;
    };
  
  if ((strstr (cmdline, "logepoch")) != NULL)
    {
      LP->logepoch = (int) value;
      return 0;
    };
  
  if ((strstr (cmdline, "batch")) != NULL)
    {
      if (!(int) value)
	LP->batch = 0;
      else
	LP->batch = 1;
      return 0;
    };
  
  if ((strstr (cmdline, "at05")) != NULL)
    {
      if (!(int) value)
	LP->at05 = 0;
      else
	LP->at05 = 1;
      return 0;
    };

  if ((strstr (cmdline, "nopass")) != NULL)
    {
      if (!(int) value)
	LP->nopass = 0;
      else
	LP->nopass = 1;
      return 0;
    };
  
  
  if ((strstr (cmdline, "asym")) != NULL)
    {
      if (!(int) value)
	LP->asym = 0;
      else
	LP->asym = 1;
      return 0;
    };
  
  if ((strstr (cmdline, "netfile")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->netfilename, S);
	  return 0;
	}
      else return -1;
    }
       
  if ((strstr (cmdline, "logfile")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->logfilename, S);
	  return 0;
	}
      else return -1;
    };

  if ((strstr (cmdline, "gnufile")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->gnufilename, S);
	  return 0;
	}
      else return -1;
    };

  if ((strstr (cmdline, "trainingfile")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->trainingfilename, S);
	  return 0;
	}
      else return -1;
    };

  if ((strstr (cmdline, "generalizationfile")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->generalizationfilename, S);
	  return 0;
	}
      else return -1;
    };
 
  if ((strstr (cmdline, "testfile")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->testfilename, S);
	  return 0;
	}
      else return -1;
    };

  if ((strstr (cmdline, "loadnet")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->loadnetfilename, S);
	  return 0;
	}
      else return -1;
    };

  if ((strstr (cmdline, "savenet")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->savenetfilename, S);
	  return 0;
	}
      else return -1;
    };

  if ((strstr (cmdline, "trainresult")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->trainresultname, S);
	  return 0;
	}
      else return -1;
    };

  if ((strstr (cmdline, "genresult")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->genresultname, S);
	  return 0;
	}
      else return -1;
    };

  if ((strstr (cmdline, "testresult")) != NULL)
    {
      sscanf (cmdline, "%s %s", S, S);
      if (strlen (S) != 0)
	{
	  strcpy (LP->testresultname, S);
	  return 0;
	}
      else return -1;
    };


  /* if we have not already left, then the command is invalid */
  return -1;
}; /* Process_Cmd */


/* Read_Commands: read global variables from control file */
int Read_Commands (const char *filename, LParam *LP)
{
 FILE *cmdfile;
 char cmdline[80];
 int i;

 if ((cmdfile = fopen (filename, "r")) == NULL)
   {
     printf ("Cannot open command file %s.\n", filename);
     return -1;
   };

 printf ("Reading commands from ini-file %s.\n\n", filename);
 i=0;       
 while (!feof (cmdfile))
   {
     fgets (cmdline, 79, cmdfile);
     i++;
     Remove_Whitespace (cmdline);
     if ((cmdline[0] != '#') && (cmdline[0] != '%') && (strlen(cmdline) != 0))
       if (Process_Cmd (cmdline, LP) < 0)
	 printf ("Error in %s line %i: %s\n", filename, i, cmdline);
   };

 fclose (cmdfile);
 return 0;
}; /* Read_Commands */

void Print_Parameters (int nin, int nout, LParam *LP)
{
  int i;

  printf("Parameters:\n");
  printf("===========\n");
  printf("Type of fuzzy system = %i\n", LP->fstype);
  printf("Input MSF: ");
  for (i=0; i<nin; i++)
    printf("%i:%i ", LP->imsf[i], LP->iform[i]);  
  printf("\nOutput MSF: ");
  for (i=0; i<nout; i++)
    printf("%i:%i ", LP->omsf[i], LP->oform[i]);
  printf ("\ntskinit = %lf\n", LP->tskinit);
  printf("outstretch = %lf\n", LP->outstretch);
  printf("t-norm = %i\n", LP->tnorm);
  printf("t-conorm = %i\n", LP->snorm);
  printf("defuzzification = %i\n", LP->defuzz);
  printf("sigma_a = %lf\n", LP->sigma_a);
  printf("sigma_b = %lf\n", LP->sigma_b);
  printf("sigma_c = %lf\n", LP->sigma_c);
  printf("sigma_d = %lf\n", LP->sigma_d);
  printf("sigma_t = %lf\n", LP->sigma_t);
  printf("momentum = %lf\n", LP->momentum);
  printf("maxrules = %i\n", LP->maxrules);
  printf("percent = %lf\n", LP->percent);
  printf("minerror = %lf\n", LP->minerror);
  printf("chgepoch = %i\n", LP->chgepoch);
  printf("minepoch = %i\n", LP->minepoch);
  printf("maxepoch = %i\n", LP->maxepoch);
  printf("logepoch = %i\n", LP->logepoch);
  printf("batch = %i\n", LP-> batch);
  printf("at05 = %i\n", LP->at05);
  printf("nopass = %i\n", LP->nopass);
  printf("asym = %i\n", LP->asym);
  printf("netfile = %s\n", LP->netfilename);
  printf("load net from = %s\n", LP->loadnetfilename);
  printf("logfile = %s\n", LP->logfilename);
  printf("training patterns = %s\n", LP->trainingfilename);
  printf("training result = %s\n", LP->trainresultname);
  printf("generalization patterns = %s\n", LP->generalizationfilename);
  printf("generalization result = %s\n", LP->genresultname);
  printf("test patterns = %s\n", LP->testfilename);
  printf("test result = %s\n", LP->testresultname);
  
}; /* Print_Parameters */

int Write_MFuzzy_System (const char *filename, MFuzzySystem *M)
{
  int     i,j;
  FILE    *netfile;
  LingVar *lv;
  FSet     mu;
  int     tnormflag;
  int     snormflag;
  int     defuzzflag;

  if ((netfile = fopen (filename, "w")) == NULL)
    {
      printf ("Cannot create network file %s.\n", filename);
      return -1;
    };

  fprintf (netfile, "# Network file of NFIDENT\n\n");

  fprintf (netfile, "# Type of Fuzzy System (0 = Mamadani, 1 = TSK)\n");
  fprintf (netfile, "TYPE\n");
  fprintf (netfile, "0\n\n");

  fprintf (netfile, "# Dimensions of the fuzzy system: ");
  fprintf (netfile, "inputs, rules, outputs\n");
  fprintf (netfile, "DIMENSIONS\n");
  fprintf (netfile, "%i %i %i\n\n", M->in, M->rules, M->out);
  
  fprintf (netfile, "# T-norm to use with the fuzzy system (0 = min, ");
  fprintf (netfile, "1 = product).\n");
  fprintf (netfile, "T-NORM\n");
  if (M->tnorm == prod) 
    tnormflag = 1;
  else tnormflag = 0;
  fprintf (netfile, "%i\n\n", tnormflag);

  fprintf (netfile, "# T-conorm to use with the fuzzy system (0 = max)\n");
  fprintf (netfile, "T-CONORM\n");
  if (M->snorm == maximum) 
    snormflag = 0;
  else snormflag = 0;
  fprintf (netfile, "%i\n\n", snormflag);

  fprintf (netfile, "# Defuzzyification procedure (0 = COG, 1 = MOM)\n");
  fprintf (netfile, "DEFUZZIFICATION\n");
  if (M->defuzz == MOM) 
    defuzzflag = 1;
  else defuzzflag = 0;
  fprintf (netfile, "%i\n\n", defuzzflag);

  fprintf (netfile, "# Parameters of the fuzzy sets for all variables\n");
  fprintf (netfile, "# For each variable: number of fuzzy sets, lower and\n");
  fprintf (netfile, "# upper bounds, name, then a row for each fuzzy set.\n");
  fprintf (netfile, "# Parameters:  fstype, a, b, c, d, ls, rs, name\n");

  fprintf (netfile, "# First input then output variables.\n");
  fprintf (netfile, "VARIABLES\n");
  for (i = 0; i < M->in; i++)
    {
      lv = M->invars[i];
      fprintf (netfile, "%i %lf %lf %s\n", lv->count, lv->lower, lv->upper,
	       lv->name);
      for (j = 0; j < lv->count; j++)
	{
	  mu = lv->fsets[j];
	  fprintf (netfile, "%i %lf %lf %lf %lf %i %i %s\n",
		   mu.form, mu.a, mu.b, mu.c, mu.d, mu.ls, mu.rs, mu.name); 
	};
      fprintf (netfile, "\n");
    };

  for (i = 0; i < M->out; i++)
    {
      lv = M->outvars[i];
      fprintf (netfile, "%i %lf %lf %s\n", lv->count, lv->lower, lv->upper,
	       lv->name);
      for (j = 0; j < lv->count; j++)
	{
	  mu = lv->fsets[j];
	  fprintf (netfile, "%i %lf %lf %lf %lf %i %i %s\n",
		   mu.form, mu.a, mu.b, mu.c, mu.d, mu.ls, mu.rs, mu.name); 
	};
      fprintf (netfile, "\n");
    };

  fprintf (netfile, "\n# Rule base of the fuzzy system.\n");
  fprintf (netfile, "# One row for each rule, the entries are indices of\n");
  fprintf (netfile, "# fuzzy sets (counting from 0, -1 means the ");
  fprintf (netfile, "respective variable is not used).\n");
  fprintf (netfile, "\n\n");

  fprintf (netfile, "RULES\n");
  for (i = 0; i < M->rules; i++)
    {
      for (j = 0; j < M->in; j++)
	fprintf (netfile, "%i ", M->rulebase[i]->antecedent[j]);
      fprintf (netfile, "   ");
      for (j = 0; j < M->out; j++)
	fprintf (netfile, "%i ", M->rulebase[i]->conclusion[j]);
      fprintf (netfile, "\n");
    };

  fprintf (netfile, "\n# End of NFIDENT network file\n");
  fclose (netfile);
  return 0;
}; /* Write MFuzzy System */


void Get_Input_Line (FILE *file, char *line)
{
  do
    {
      fgets (line, 79, file);
      Remove_Whitespace (line);
    }
  while ((line[0] == '#') || (line[0] == '%') || (line[0] == '\0'));

}; /* Get Input Line */

MFuzzySystem *Read_MFuzzy_System(const char *filename, LParam *LP)
{
  MFuzzySystem *M;
  FILE *netfile;
  char line[80];
  int nin, nout, nrules;
  int fstype, flag;
  int i,j;

  if ((netfile = fopen (filename, "r")) == NULL)
    {
      printf ("Cannot open network file %s.\n", filename);
      return NULL;
    };

  /* check type of fuzzy system */
  Get_Input_Line (netfile, line);
  if (strstr (line,"TYPE") == NULL)
    {
      printf ("Error in file in %s looking for TYPE.\n", filename);
      return (NULL);
    };
  
  fscanf (netfile, "%i", &fstype);
  
  if (fstype != MAMDANI)
    {
      printf ("Wrong type of fuzzy system in %s.\n", filename);
      return (NULL);
    };

  /* read dimensions of fuzzy system */
  Get_Input_Line (netfile, line);
  if (strstr(line,"DIMENSION") == NULL)
    {
      printf ("Error in file %s looking for DIMENSIONS.\n", filename);
      return (NULL);
    };

  /* be careful, these values are currently unchecked! */
  fscanf (netfile, "%i %i %i", &nin, &nrules, &nout);

  /* create fuzzy system */   
  if ((M = Create_Mamdani_Fuzzy_System (nin, nout, LP)) == NULL)
    {
      printf ("Failed to create Mamdani fuzzy system.\n");
      return NULL;
    };

  M->rules = nrules;

  /* read t-norm flag */
  Get_Input_Line (netfile, line);
  if (strstr(line,"T-NORM") == NULL)
    {
      printf ("Error in file %s looking for T-NORM.\n", filename);
      return (NULL);
    };

  /* be careful, these values are currently unchecked! */
  fscanf (netfile, "%i", &flag);
  if (flag == 1)
    M->tnorm = prod;
  else M->tnorm = minimum;

  /* read t-conorm flag */
  Get_Input_Line (netfile, line);
  if (strstr(line,"T-CONORM") == NULL)
    {
      printf ("Error in file %s looking for T-CONORM.\n", filename);
      return (NULL);
    };

  /* be careful, these values are currently unchecked! */
  fscanf (netfile, "%i", &flag);
  if (flag == 0)
    M->snorm = maximum;
  else M->snorm = maximum;

  /* read defuzzification flag */
  Get_Input_Line (netfile, line);
  if (strstr(line,"DEFUZZIFICATION") == NULL)
    {
      printf ("Error in file %s looking for DEFUZZIFICATION.\n", filename);
      return (NULL);
    };

  /* be careful, these values are currently unchecked! */
  fscanf (netfile, "%i", &flag);
  if (flag == 1)
    M->defuzz = MOM;
  else M->defuzz = COG;

  Get_Input_Line (netfile, line);
  if (strstr(line,"VARIABLES") == NULL)
    {
      printf ("Error in file %s looking for VARIABLES.\n", filename);
      return (NULL);
    };
     
  for (i = 0; i < nin; i++)
    {/* read the input variables */
      if ((M->invars[i] = (LingVar *) calloc (1, sizeof(LingVar))) == NULL)
	{
	  printf ("Failed to create %ith input variable.\n", i);
	  return NULL;
	};
      /* be careful, these values are currently unchecked! */
      fscanf (netfile, "%i %lf %lf %19s", &(M->invars[i]->count), 
	      &(M->invars[i]->lower), &(M->invars[i]->upper), 
	      M->invars[i]->name);
      for (j = 0; j < M->invars[i]->count; j++)
	{ /* read the fuzzy set parameters */
	  /* be careful, these values are currently unchecked! */
	  fscanf (netfile, "%i %lf %lf %lf %lf %i %i %19s",
		  &(M->invars[i]->fsets[j].form), &(M->invars[i]->fsets[j].a),
		  &(M->invars[i]->fsets[j].b), &(M->invars[i]->fsets[j].c),
		  &(M->invars[i]->fsets[j].d), &(M->invars[i]->fsets[j].ls),
		  &(M->invars[i]->fsets[j].rs), M->invars[i]->fsets[j].name);
	}; /* for j */
    }; /* for i */

  for (i = 0; i < nout; i++)
    {/* read the output variables */
      if ((M->outvars[i] = (LingVar *) calloc (1, sizeof(LingVar))) == NULL)
	{
	  printf ("Failed to create %ith output variable.\n", i);
	  return NULL;
	};
      /* be careful, these values are currently unchecked! */
      fscanf (netfile, "%i %lf %lf %19s", &(M->outvars[i]->count), 
	      &(M->outvars[i]->lower), &(M->outvars[i]->upper), 
	      M->outvars[i]->name);
      for (j = 0; j < M->outvars[i]->count; j++)
	{ /* read the fuzzy set parameters */
	  /* be careful, these values are currently unchecked! */
	  fscanf (netfile, "%i %lf %lf %lf %lf %i %i %19s",
		  &(M->outvars[i]->fsets[j].form), 
		  &(M->outvars[i]->fsets[j].a),
		  &(M->outvars[i]->fsets[j].b), &(M->outvars[i]->fsets[j].c),
		  &(M->outvars[i]->fsets[j].d), &(M->outvars[i]->fsets[j].ls),
		  &(M->outvars[i]->fsets[j].rs), M->outvars[i]->fsets[j].name);
	}; /* for j */
    }; /* for i */

  /* read rule base */
  Get_Input_Line (netfile, line);
  if (strstr(line,"RULES") == NULL)
    {
      printf ("Error in file %s looking for RULES.\n", filename);
      return (NULL);
    };
  for (i = 0; i < nrules; i++)
    {
      /* allocate a new rule, and set antecedent and conclusion */
      if ((M->rulebase[i] = Allocate_New_Mamdani_Rule(M->in, M->out)) == NULL)
	{
	  printf ("Cannot allocate memory for rule %i.\n", i);
	  return NULL;
	};

      for (j = 0; j < M->in; j++)
	{
	  fscanf (netfile, "%i", &(M->rulebase[i]->antecedent[j]));
	};
      for (j = 0; j < M->out; j++)
	{
	  fscanf (netfile, "%i", &(M->rulebase[i]->conclusion[j]));
	};
    };
  fclose (netfile);
  return M;
}; /* Read_MFuzzy_System */


int Write_TSKFuzzy_System (const char *filename, TSKFuzzySystem *TSK)
{
  int     i,j,k;
  FILE    *netfile;
  LingVar *lv;
  FSet     mu;
  int     tnormflag;


  if ((netfile = fopen (filename, "w")) == NULL)
    {
      printf ("Cannot create network file %s.\n", filename);
      return -1;
    };

  fprintf (netfile, "# Network file of NFIDENT\n\n");

  fprintf (netfile, "# Type of Fuzzy System (0 = Mamadani, 1 = TSK)\n");
  fprintf (netfile, "TYPE\n");
  fprintf (netfile, "1\n\n");

  fprintf (netfile, "# Dimensions of the fuzzy system: ");
  fprintf (netfile, "inputs, rules, outputs\n");
  fprintf (netfile, "DIMENSIONS\n");
  fprintf (netfile, "%i %i %i\n\n", TSK->in, TSK->rules, TSK->out);
  
  fprintf (netfile, "# T-norm to use with the fuzzy system (0 = min, ");
  fprintf (netfile, "1 = product).\n");
  fprintf (netfile, "T-NORM\n");
  if (TSK->tnorm == prod) 
    tnormflag = 1;
  else tnormflag = 0;
  fprintf (netfile, "%i\n\n", tnormflag);

  fprintf (netfile, "# Parameters of the fuzzy sets for all variables\n");
  fprintf (netfile, "# For each variable: number of fuzzy sets, lower and\n");
  fprintf (netfile, "# upper bounds, name, then a row for each fuzzy set.\n");
  fprintf (netfile, "# Parameters:  fstype, a, b, c, d, ls, rs, name\n");

  fprintf (netfile, "# A TSK system has only fuzzy sets for the inputs.\n");
  fprintf (netfile, "VARIABLES\n");
  for (i = 0; i < TSK->in; i++)
    {
      lv = TSK->invars[i];
      fprintf (netfile, "%i %lf %lf %s\n", lv->count, lv->lower, lv->upper,
	       lv->name);
      for (j = 0; j < lv->count; j++)
	{
	  mu = lv->fsets[j];
	  fprintf (netfile, "%i %lf %lf %lf %lf %i %i %s\n",
		   mu.form, mu.a, mu.b, mu.c, mu.d, mu.ls, mu.rs, mu.name); 
	};
      fprintf (netfile, "\n");
    };

  fprintf (netfile, "\n# Rule base of the fuzzy system.\n");
  fprintf (netfile, "# OUT + 1 rows for each rule, the entries of row 1\n");
  fprintf (netfile, "# are indices of fuzzy sets (counting from 0, -1\n");
  fprintf (netfile, "# means the respective variable is not used).\n");
  fprintf (netfile, "# The entries of the other OUT rows are coefficients\n");
  fprintf (netfile, "# (one row for each output variable, each row has one\n");
  fprintf (netfile, "# number for each input variable and a constant).\n");
  fprintf (netfile, "\n\n");

  fprintf (netfile, "RULES\n");
  for (i = 0; i < TSK->rules; i++)
    {
      for (j = 0; j < TSK->in; j++)
	{
	  fprintf (netfile, "%i ", TSK->rulebase[i]->antecedent[j]);
	};
      fprintf (netfile, "\n");
      for (j = 0; j < TSK->out; j++)
	{
	  for (k = 0; k < TSK->in+1; k++)
	    {
	      fprintf (netfile, "%lf ", TSK->rulebase[i]->conclusion[j][k]);
	    };
	  fprintf (netfile, "\n");
	};
    };

  fprintf (netfile, "\n# End of NFIDENT network file\n");
  fclose (netfile);
  return 0;
}; /* Write TSKFuzzy System */

TSKFuzzySystem *Read_TSKFuzzy_System(const char *filename, LParam *LP)
{
  TSKFuzzySystem *TSK;
  FILE *netfile;
  char line[80];
  int nin, nout, nrules;
  int fstype, flag;
  int i,j,k;

  if ((netfile = fopen (filename, "r")) == NULL)
    {
      printf ("Cannot open network file %s.\n", filename);
      return NULL;
    };

  /* check type of fuzzy system */
  Get_Input_Line (netfile, line);
  if (strstr (line,"TYPE") == NULL)
    {
      printf ("Error in file in %s looking for TYPE.\n", filename);
      return (NULL);
    };
  
  fscanf (netfile, "%i", &fstype);
  
  if (fstype != SUGENO)
    {
      printf ("Wrong type of fuzzy system in %s.\n", filename);
      return (NULL);
    };

  /* read dimensions of fuzzy system */
  Get_Input_Line (netfile, line);
  if (strstr(line,"DIMENSION") == NULL)
    {
      printf ("Error in file %s looking for DIMENSIONS.\n", filename);
      return (NULL);
    };

  /* be careful, these values are currently unchecked! */
  fscanf (netfile, "%i %i %i", &nin, &nrules, &nout);

  /* create fuzzy system */   
  if ((TSK = Create_TSK_Fuzzy_System (nin, nout, LP)) == NULL)
    {
      printf ("Failed to create Takagi/Sugeno/Kang fuzzy system.\n");
      return NULL;
    };

  TSK->rules = nrules;

  /* read t-norm flag */
  Get_Input_Line (netfile, line);
  if (strstr(line,"T-NORM") == NULL)
    {
      printf ("Error in file %s looking for T-NORM.\n", filename);
      return (NULL);
    };

  /* be careful, these values are currently unchecked! */
  fscanf (netfile, "%i", &flag);
  if (flag == 1)
    TSK->tnorm = prod;
  else TSK->tnorm = minimum;

  Get_Input_Line (netfile, line);
  if (strstr(line,"VARIABLES") == NULL)
    {
      printf ("Error in file %s looking for VARIABLES.\n", filename);
      return (NULL);
    };
     
  for (i = 0; i < nin; i++)
    {/* read the input variables */
      if ((TSK->invars[i] = (LingVar *) calloc (1, sizeof(LingVar))) == NULL)
	{
	  printf ("Failed to create %ith input variable.\n", i);
	  return NULL;
	};
      /* be careful, these values are currently unchecked! */
      fscanf (netfile, "%i %lf %lf %19s", &(TSK->invars[i]->count), 
	      &(TSK->invars[i]->lower), &(TSK->invars[i]->upper), 
	      TSK->invars[i]->name);
      for (j = 0; j < TSK->invars[i]->count; j++)
	{ /* read the fuzzy set parameters */
	  /* be careful, these values are currently unchecked! */
	  fscanf (netfile, "%i %lf %lf %lf %lf %i %i %19s",
		  &(TSK->invars[i]->fsets[j].form), 
		  &(TSK->invars[i]->fsets[j].a),
		  &(TSK->invars[i]->fsets[j].b), 
		  &(TSK->invars[i]->fsets[j].c),
		  &(TSK->invars[i]->fsets[j].d), 
		  &(TSK->invars[i]->fsets[j].ls),
		  &(TSK->invars[i]->fsets[j].rs), 
		  TSK->invars[i]->fsets[j].name);
	}; /* for j */
    }; /* for i */


  /* read rule base */
  Get_Input_Line (netfile, line);
  if (strstr(line,"RULES") == NULL)
    {
      printf ("Error in file %s looking for RULES.\n", filename);
      return (NULL);
    };
  for (i = 0; i < nrules; i++)
    {
      /* allocate a new rule, and set antecedent and conclusion */
      if ((TSK->rulebase[i] = Allocate_New_TSK_Rule(TSK->in, TSK->out)) 
	  == NULL)
	{
	  printf ("Cannot allocate memory for rule %i.\n", i);
	  return NULL;
	};

      for (j = 0; j < TSK->in; j++)
	{
	  fscanf (netfile, "%i", &(TSK->rulebase[i]->antecedent[j]));
	};
      for (j = 0; j < TSK->out; j++)
	{
	  for (k = 0; k < TSK->in+1; k++)
	    {
	      fscanf (netfile, "%lf", &(TSK->rulebase[i]->conclusion[j][k]));
	    };
	};
    };
  fclose (netfile);
  return TSK;
}; /* Read_TSKFuzzy_System */




/* Read_Files calls the functions for reading parameters, data and an  */
/* initial network. It takes the name of an commandfile and a parameter */
/* structure and returns a 0, if everything is fine, a -1 else */
/* Read_Files changes the global variables for networks and data. */
int Read_Files (const char *commandfile, LParam *LP)
{
  int i;

  if (0 != Read_Commands (commandfile, LP))
    {
      return -1;
    };

  if (LP->trainingfilename[0] != '\0')
    {

      if (NULL == (TrainingPattern = (PSet *) calloc (1, sizeof (PSet))))
	{
	  printf ("Error when allocating training data set!\n");
	  return -1;
	};

      if (Read_Pattern (LP->trainingfilename,  TrainingPattern) == 0)
	{
	  printf ("Training data in %s: %i patterns (in: %i, out: %i).\n",
		  LP->trainingfilename,
		  TrainingPattern->count, TrainingPattern->in, 
		  TrainingPattern->out);
	  for (i = 0; i < TrainingPattern->in + TrainingPattern->out; i++)
	    {
	      printf ("Feature %i in [%.2f,%.2f]\n", i, 
		      TrainingPattern->pmin[i],
		      TrainingPattern->pmax[i]);
	    };
	}
      else
	{
	  printf ("Error when reading patterns from %s!\n", 
		  LP->trainingfilename);
	  return -1;
	};
    }
  else
    {
      printf ("Error: A training data file must be specified!\n");
      return -1;
    };

  if (LP->generalizationfilename[0] != '\0')
    {
      if (NULL == (GeneralizationPattern = (PSet *) calloc (1, sizeof (PSet))))
	{
	  printf ("Error when allocating generalization data set!\n");
	  return -1;
	};

      if (Read_Pattern (LP->generalizationfilename,  
			GeneralizationPattern) == 0)
	{
	  if ((TrainingPattern->in != GeneralizationPattern->in) ||
	      (TrainingPattern->out != GeneralizationPattern->out))
	    {
	      printf ("Generalization data dimensions do not fit to training data!\n");
	      return -1;
	    }
	  
	  printf ("Generalization data in %s: %i patterns (in %i, out %i).\n",
		  LP->generalizationfilename, GeneralizationPattern->count, 
		  GeneralizationPattern->in, GeneralizationPattern->out);
	  for (i = 0; i < GeneralizationPattern->in + 
		          GeneralizationPattern->out; i++)
	    {
	      printf ("Feature %i in [%.2f,%.2f]\n", i, 
		      GeneralizationPattern->pmin[i],
		      GeneralizationPattern->pmax[i]);
	    };
	}
      else
	{
	  printf ("Error when reading patterns from %s!\n", 
		  LP->generalizationfilename);
	  return -1;
	};
    };

  if (LP->testfilename[0] != '\0')
    {
      if (NULL == (TestPattern = (PSet *) calloc (1, sizeof (PSet))))
	{
	  printf ("Error when allocating test data set!\n");
	  return -1;
	};

      if (Read_Pattern (LP->testfilename,  TestPattern) == 0)
	{
	  if ((TrainingPattern->in != TestPattern->in) ||
	      (TrainingPattern->out != TestPattern->out))
	    {
	      printf ("Test data dimensions do not fit to training data!\n");
	      return -1;
	    }

	  printf ("Test data in %s: %i patterns (in: %i, out: %i).\n",
		  LP->testfilename, TestPattern->count, TestPattern->in, 
		  TestPattern->out);
	  for (i = 0; i < TestPattern->in + TestPattern->out; i++)
	    {
	      printf ("Feature %i in [%.2f,%.2f]\n", i, 
		      TestPattern->pmin[i],
		      TestPattern->pmax[i]);
	    };
	}
      else
	{
	  printf ("Error when reading patterns from %s!\n", 
		  LP->testfilename);
	  return -1;
	};
    };

  if (LP->loadnetfilename[0] != '\0')
    {
      if (LP->fstype == 0)
	{
	  printf ("Read an initial Nefprox system from %s.\n",
		  LP->loadnetfilename);
	  if ((Nefprox = Read_MFuzzy_System (LP->loadnetfilename, LP)) == NULL)
	   {
	     return -1;
	   };
	}
      else
	{
	  printf ("Read an initial ANFIS system from %s.\n",
		  LP->loadnetfilename);
	  if ((Anfis = Read_TSKFuzzy_System (LP->loadnetfilename, LP)) == NULL)
	    {
	      return -1;
	    };	  
	};
    };

  Print_Parameters (TrainingPattern->in, TrainingPattern->out, LP);
  return 0;

}; /* Read_Files */


int Build_Filename (char *filename, int n)
{
  char *rest;

  if ((rest = (strstr (filename,"*"))) != NULL)
    {
      rest[0] = '\0';
      sprintf (filename, "%s%i%s", filename, n, rest+1);
    }
}; /* Build Filename*/ 


int Test_Netfile (const char *filename)
{
  FILE *netfile;
  int fstype;
  char line[80];

  if ((netfile = fopen (filename, "r")) == NULL)
    {
      printf ("Cannot open network file %s.\n", filename);
      return -1;
    };

  /* check type of fuzzy system */
  Get_Input_Line (netfile, line);
  if (strstr (line,"TYPE") == NULL)
    {
      printf ("Error in file in %s looking for TYPE.\n", filename);
      return -1;
    };
  
  fscanf (netfile, "%i", &fstype);
  
  fclose (netfile);
  return fstype;

}; /* Test_Netfile */
 
int Read_Net_and_Data (const char *netfile, const char *datafile, LParam *LP)
{
  int fstype;
  int i;
  
  fstype = Test_Netfile (netfile);
  switch (fstype)
    {
    case MAMDANI:
      if (NULL == (Nefprox = Read_MFuzzy_System (netfile, LP)))
	{
	  return -1;
	};
      break;
    case SUGENO:
      if (NULL == (Anfis = Read_TSKFuzzy_System(netfile, LP)))
	{
	  return -1;
	};
      break;
    default:
      return -1;
    }; /* switch */

  if (NULL == (TestPattern = (PSet *) calloc (1, sizeof (PSet))))
    {
      printf ("Error when allocating test data set!\n");
      return -1;
    };

  if (Read_Pattern (datafile,  TestPattern) == 0)
    {
      printf ("Test data in %s: %i patterns (in: %i, out: %i).\n",
	      datafile, TestPattern->count, TestPattern->in, 
	      TestPattern->out);
      for (i = 0; i < TestPattern->in + TestPattern->out; i++)
	{
	  printf ("Feature %i in [%.2f,%.2f]\n", i, 
		  TestPattern->pmin[i],
		  TestPattern->pmax[i]);
	};
    }
  else
    {
      printf ("Error when reading patterns from %s!\n", 
	      datafile);
      return -1;
    };  

  return 0;
};/* Read_Net_and_Data */

int Read_Net(const char *netfile, LParam *LP)
{
  int fstype;
  
  fstype = Test_Netfile (netfile);
  switch (fstype)
    {
    case MAMDANI:
      if (NULL == (Nefprox = Read_MFuzzy_System (netfile, LP)))
	{
	  return -1;
	};
      break;
    case SUGENO:
      if (NULL == (Anfis = Read_TSKFuzzy_System(netfile, LP)))
	{
	  return -1;
	};
      break;
    default:
      return -1;
    }; /* switch */
  
  return 0;
};/* Read_Net */


