/* NFIDENT: a program for neuro-fuzzy function approximation

   Copyright (C) 1997-2000 

   Detlef Nauck (nauck@ieee.org, 
   http:\\fuzzy.cs.uni-magdeburg.de/~nauck)  
   hereby disclaims all copyright interests in the program 
   NFIDENT written by Detlef Nauck

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 1, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <math.h>
#include <stdio.h>
#include "global.h"

/* CAUTION: Does not work yet! */
/* This file contains functions for the least mean square estimation routine */
/* used by ANFIS */

/* Matrix_by_Vector multiplies a l(ines) by c(olumns) matrix M and a vector */
/* V (column) with c components. The result is written to the vector R with */
/* l components, which has to be allocated before (R = M * V).              */

void Matrix_by_Vector (double *R, matrix *M, double *V, int c, int l)
{
  int i,j;

  for (i = 0; i < l; i++)
    {
      R[i] = 0;
      for (j = 0; j < c; j++)
	{
	  R[i] += M[i][j] * V[j];
	}; /* for j */
    }; /* for i */
  return;
}; /* Matrix by Vector */

/* Matrix_by_Vector multiplies a transposed vector V (line) with l        */
/* components and a l(ines) by c(olumns) matrix M. The result is written  */
/* to the vector R with c components, which has to be allocated before    */
/* (R = V * M).                                                           */
void Vector_by_Matrix (double *R, double * V, matrix *M, int c, int l)
{
  int i,j;

  for (i = 0; i < c; i++)
    {
      R[i] = 0;
      for (j = 0; j < l; j++)
	{
	  R[i] += M[j][i] * V[j];
	}; /* for j */
    }; /* for i */
  return;
}; /* Matrix by Vector */


/* Outer_Product multiplies a vector V (column) with l components by a    */
/* vector W (line) with c components and writes the result to the matrix  */
/* M with l * c components, which has to be allocated before (M = V * W). */
void Outer_Product (matrix *M, double *V, double *W, int c, int l)
{
  int i,j;

  for (i = 0; i < l; i++)
    {
      for (j = 0; j < c; j++)
	{
	  M[i][j] = V[i] * W[j];
	}; /* for j */
    }; /* for i */
  return;
}; /* Outer_Product */

double Inner_Product (double *V, double *W, int c)
{
  int i;
  double x;

  x = 0;

  for (i = 0; i < c; i++)
    {
      x += V[i] * W[i];
    }; /* for i */

  return x;
}; /* Inner_Product */


int LMSE_Step (double *X, double *A, double error, int d, int k, int n)
{
 
 static double  *SA;
 static double  *AS;
 static matrix *SS;
 static matrix **S;
 static int    washere; /* magic number to remember that we were here before */

 int i,j;
 double y;

 if (washere != 12345)
   {
     if ((SA = (double *) (calloc (d, sizeof(double)))) == NULL)
       {
	 printf ("Cannot allocate memory for LSME procedure!\n");
	 return -1;
       };
     
     if ((AS = (double *) (calloc (d, sizeof(double)))) == NULL)
       {
	 printf ("Cannot allocate memory for LSME procedure!\n");
	 return -1;
       };
     
     if ((S = (matrix **) (calloc (n, sizeof(matrix *)))) == NULL)
       {
	 printf ("Cannot allocate memory for LSME procedure!\n");
	 return -1;
       };

     for (i = 0; i < n; i++)
       {
	 if ((S[i] = (matrix *) (calloc (d, sizeof(double *)))) == NULL)
	   {
	     printf ("Cannot allocate memory for LSME procedure!\n");
	     return -1;
	   };
     
	 for (j = 0; j < d; j++)
	   {
	     if ((S[i][j] = (double *) (calloc (d, sizeof(double)))) == NULL)
	       {
		 printf ("Cannot allocate memory for LSME procedure!\n");
		 return -1;
	       };
	     S[i][j][j] = 1000000.0;
	   };
       };

     if ((SS = (matrix *) (calloc (d, sizeof(double *)))) == NULL)
       {
	 printf ("Cannot allocate memory for LSME procedure!\n");
	 return -1;
       };
     
     for (i = 0; i < d; i++)
       {
	 if ((SS[i] = (double *) (calloc (d, sizeof(double)))) == NULL)
	   {
	     printf ("Cannot allocate memory for LSME procedure!\n");
	     return -1;
	   };
       };
     washere = 12345;
   };

 Matrix_by_Vector (SA, S[k], A, d, d);
 Vector_by_Matrix  (AS, A, S[k], d, d);
 Outer_Product     (SS, SA, AS, d, d);
 y = 1 + Inner_Product (A, AS, d);
 
 for (i = 0; i < d; i++)
   {
     for (j = 0; j < d; j++)
       {
	 S[k][i][j] -= SS[i][j] / y; 
       };
   };

 Matrix_by_Vector (SA, S[k], A, d, d);

 for (i = 0; i < d; i++)
   {
     X[i] += SA[i] * error;
   };

}; /* LSME Step */



